# psim を利用するためのおまじない -- ここから
import winreg, sys, os
reg = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\Mathematical Systems Inc.\\S-Quattro Simulation System\\INFO")
rootdir = winreg.QueryValueEx(reg, "ROOTDIR")[0]
os.environ["PATH"] = rootdir + ";" + os.environ["PATH"]
sys.path.insert(0, os.path.join(rootdir, "lib3"))
# psim を利用するためのおまじない -- ここまで

from psim import *
import matplotlib.pyplot as plt
import matplotlib
import pandas as pd
import bisect
import pdb

initialize()

class Agent:
    def __init__(self, state):
        # 状態は "S"/"I"/"R"
        self.state = state
    def step(self, I, beta, gamma):
        # I は、感染者数
        if self.state == "S":
            # 感受性者は、beta * I の確率で感染する
            if np.random.uniform() <= beta * I:
                self.state = "I"
        elif self.state == "I":
            # 感染者は、gamma の確率で回復もしくは隔離
            if np.random.uniform() <= gamma:
                self.state = "R"

# エージェントの初期化
agents = []

N = 1000

for i in range(N - 10):
    agents.append(Agent("S"))
for i in range(10):
    agents.append(Agent("I"))
for i in range(0):
    agents.append(Agent("R"))

hist = []

# n 人を人数分布(dist)に従ってグループに分割する。
# 所属グループIDのリストを返す。
def makeGroup(dist, n):
    keys = np.array(list(dist.keys()))
    values = np.array(list(dist.values()))
    groups = []
    k = 0
    while len(groups) < n:
        r = np.random.uniform(0, values.sum())
        index = bisect.bisect_left(values.cumsum(), r)
        groups.extend([k] * keys[index])
        k += 1
    groups = np.array(groups[:N])
    np.random.shuffle(groups)
    return groups

# 全エージェントをグループに分割したもの
# 感染はグループ内のみで発生する
class AgentGroup:
    def __init__(self, dist, beta, gamma):
        self.dist = dist
        self.beta = beta
        self.gamma = gamma
        self.ids = makeGroup(dist, N)
        self.groups = [[] for i in range(self.ids.max() + 1)]
        for i, a in enumerate(agents):
            self.groups[self.ids[i]].append(a)
    def step(self):
        S = 0
        I = 0
        R = 0
        # 各グループに対し
        for agents in self.groups:
            # グループ内の感染者数
            i = len([a for a in agents if a.state == "I"])
            # グループ内の全エージェントを状態遷移させる
            for k in range(len(agents)):
                agents[k].step(i, self.beta, self.gamma)
            # グループ内のエージェントの状態をカウントする
            for a in agents:
                if a.state == "S":
                    S += 1
                elif a.state == "I":
                    I += 1
                elif a.state == "R":
                    R += 1
        # 記録
        hist.append([S, I, R])
        return S, I, R
        
# 世帯ごとのグループ
familyGroup = AgentGroup(
    # 世帯人員分布 {世帯人員: 重み, ...}
    {1: 35.8, 2: 29.6, 3: 16.7, 4: 12.4, 5: 5.5},
    # 単位時間あたりの感染率
    0.06,
    # 単位時間あたりの回復や隔離による除去率
    0.2)
# 移動時のグループ
vehicleGroup = AgentGroup(
    # 移動時の乗り物の乗車人数分布 {乗車人数: 重み, ...}
    {1: 20.0, 5: 30.0, 20: 30.0, 30: 20.0},
    # 単位時間あたりの感染率
    0.04,
    # 単位時間あたりの回復や隔離による除去率
    0.2)
# 職場グループ
officeGroup = AgentGroup(
    # 日中の職場(学校)の人数 {乗車人数: 重み, ...}
    {1: 20.0, 5: 30.0, 20: 30.0, 30: 20.0},
    # 単位時間あたりの感染率
    0.03,
    # 単位時間あたりの回復や隔離による除去率
    0.2)

def proc():
    while True:
        # 世帯での感染
        for k in range(4):
            S, I, R = familyGroup.step()
            yield pause(1)
        # 出社移動時の感染
        for k in range(1):
            S, I, R = vehicleGroup.step()
            yield pause(1)
        # 日中の職場での感染
        for k in range(4):
            S, I, R = officeGroup.step()
            yield pause(1)
        # 帰宅移動時の感染
        for k in range(1):
            S, I, R = vehicleGroup.step()
            yield pause(1)

activate(proc)()

start(until = 100)

hist = np.array(hist)
plt.plot(hist[:, 0], label = "S")
plt.plot(hist[:, 1], label = "I")
plt.plot(hist[:, 2], label = "R")
plt.legend()
plt.show()
#plt.savefig("psim9-sir.png")

